/* wipe.c - writes 0's over files and optionally removes them
 * 
 * Calvin Clark <ckclark@mit.edu>
 *
 * Compile with -DBSD on old BSD systems, and -DNO_FSYNC
 * if you don't have fsync().  
 *
 * Otherwise, the code should work on most POSIX/ANSI environments.
 */

/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#ifndef BSD
#include <unistd.h>
#include <stdlib.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef BSD
#ifndef S_ISREG
#define S_ISREG(mode) 	(((mode) & S_IFMT) == S_IFREG)
#endif
extern int errno;
extern char *sys_errlist[];
#define strerror(code)  sys_errlist[(code)]
#endif

static char *whoami;
static char version[] = "1.0beta-p1";

/* This function is taken from the GNU fileutils 2.5 */

int
yesno ()
{
  int c;
  int rv;

  fflush (stderr);
  c = getchar ();
  rv = (c == 'y') || (c == 'Y');
  while (c != EOF && c != '\n')
    c = getchar ();

  return rv;
}

int
wipe_file (filename, rm, query)
     char *filename;		/* file to wipe */
     int rm;			/* non-zero means remove file */
     int query;			/* non-zero means ask before wiping file */
{
  int fd, ret_val, i;
  extern int errno;
  char buf[BUFSIZ];
  struct stat statbuf;
  int nbufs, leftover;

  /* get a bunch of zeros */
#ifdef BSD
  bzero (buf, BUFSIZ);
#else
  memset (buf, 0, BUFSIZ);
#endif
  /* find out what kind of file it is and how big it is */

  ret_val = lstat (filename, &statbuf);
  if (ret_val < 0)
    {
      fprintf (stderr, "%s: %s: %s\n", whoami, filename,
	       strerror (errno));
      return 1;
    }
  if (!S_ISREG (statbuf.st_mode))
    {
      fprintf (stderr, "%s: %s is not a regular file!\n", whoami,
	       filename);
      return 1;
    }
  nbufs = statbuf.st_size / BUFSIZ;
  leftover = statbuf.st_size % BUFSIZ;

  /* ask user before trying to open file */

  if (query)
    {
      printf ("wipe `%s'? ", filename);
      if (!yesno ())
	return 0;		/* not a failure */
    }

  fd = open (filename, O_WRONLY, 0);
  if (fd < 0)
    {
      fprintf (stderr, "%s: error opening file %s: %s\n",
	       whoami, filename, strerror (errno));
      return 1;
    }

  /* start writing nuls over the file in chunks of BUFSIZ length */

  for (i = 0; i < nbufs; i++)
    {
      ret_val = write (fd, buf, BUFSIZ);
      if (ret_val != BUFSIZ)
	{
	  fprintf (stderr, "%s: error writing to file %s: %s\n",
		   whoami, filename, strerror (errno));
	  (void) close (fd);
	  return 1;
	}
    }

  /* do what's left */

  ret_val = write (fd, buf, leftover);

  if (ret_val != leftover)
    {
      fprintf (stderr, "%s: error writing to file %s: %s\n",
	       whoami, filename, strerror (errno));
      (void) close (fd);
      return 1;
    }

  /*
   * On AFS, it is important to insure that the
   * file is written from the client to the server,
   * now that it is zero'd.  We perform and fsnyc()
   * and a close() and check return values
   * to insure that we didn't go over quota or
   * something.  The OS/2 system I tried doesn't
   * have fsync() (perhaps all writes are synchronous
   * on OS/2?)
   */
#ifndef NO_FSYNC
  ret_val = fsync (fd);
  if (ret_val < 0)
    {
      fprintf (stderr, "%s: fsync() error on %s: %s\n",
	       whoami, filename, strerror (errno));
      (void) close (fd);
      return 1;
    }
#endif
  ret_val = close (fd);
  if (ret_val < 0)
    {
      fprintf (stderr, "%s: close() error on %s: %s\n",
	       whoami, filename, strerror (errno));
      return 1;
    }

  /* actually remove the file if requested */

  if (rm)
    {
      ret_val = unlink (filename);
      if (ret_val < 0)
	{
	  fprintf (stderr, "%s: error unlinking file %s: %s\n",
		   whoami, filename, strerror (errno));
	  return 1;
	}
    }
  return 0;
}

void
main (argc, argv)
     int argc;
     char **argv;
{
  char *filename;
  int exit_status = 0;
  int rm = 0;
  int interactive = 0;
  int c;

  extern int optind;
  extern int getopt ();

  whoami = argv[0];

  if (argc < 2)
    {
      fprintf (stderr, "Usage: %s [-i] [-d] [-help] [-v] <file1> <file2> ...\n",
	       whoami);
      exit (1);
    }
  while ((c = getopt (argc, argv, "idhv")) != EOF)
    {
      switch (c)
	{
	case 'i':
	  interactive = 1;
	  break;
	case 'd':
	  rm = 1;
	  break;
	case 'h':
	  fprintf (stderr, "Usage: %s [-i] [-d] [-help] [-v] <file1> <file2> ...\n",
		   whoami);
	  fprintf (stderr, "\t-d means actually delete the file(s)\n");
	  fprintf (stderr, "\t-i means ask before wiping each file\n");
	  fprintf (stderr, "\t-help prints this message\n");
          fprintf (stderr, "\t-v prints the version\n");
	  exit (0);
	case 'v':
          printf ("wipe version %s\n", version);
	  break;
	default:
	  fprintf (stderr, "Usage: %s [-i] [-d] [-help] <file1> <file2> ...\n",
		   whoami);
	  exit (1);
	}
    }

  for (; optind < argc; optind++)
    {
      filename = argv[optind];
      exit_status |= wipe_file (filename, rm, interactive);
    }

  exit (exit_status);
}
